@@ -17,7 +17,7 @@
 %                                 June 2001                                   %
 %                                                                             %
 %                                                                             %
-%  Copyright 1999-2019 ImageMagick Studio LLC, a non-profit organization      %
+%  Copyright 1999-2020 ImageMagick Studio LLC, a non-profit organization      %
 %  dedicated to making software imaging solutions freely available.           %
 %                                                                             %
 %  You may not use this file except in compliance with the License.  You may  %
@@ -478,10 +478,21 @@ static Image *ReadJP2Image(const ImageInfo *image_info,ExceptionInfo *exception)
           pixel,
           scale;
 
+        ssize_t data_index = y/jp2_image->comps[i].dy*
+                             image->columns/jp2_image->comps[i].dx+
+                             x/jp2_image->comps[i].dx;
+        if(data_index < 0 ||
+           data_index >= (jp2_image->comps[i].h * jp2_image->comps[i].w))
+        {
+          opj_destroy_codec(jp2_codec);
+          opj_image_destroy(jp2_image);
+          ThrowReaderException(CoderError,
+                               "IrregularChannelGeometryNotSupported")
+        }
         scale=QuantumRange/(double) ((1UL << jp2_image->comps[i].prec)-1);
-        pixel=scale*(jp2_image->comps[i].data[y/jp2_image->comps[i].dy*
-          image->columns/jp2_image->comps[i].dx+x/jp2_image->comps[i].dx]+
-          (jp2_image->comps[i].sgnd ? 1UL << (jp2_image->comps[i].prec-1) : 0));
+        pixel=scale*(jp2_image->comps[i].data[data_index] +
+              (jp2_image->comps[i].sgnd ?
+               1UL << (jp2_image->comps[i].prec-1) : 0));
         switch (i)
         {
            case 0:
@@ -870,6 +881,7 @@ static MagickBooleanType WriteJP2Image(const ImageInfo *image_info,Image *image)
     {
       parameters->tcp_distoratio[0]=(double) image_info->quality;
       parameters->cp_fixed_quality=OPJ_TRUE;
+      parameters->cp_disto_alloc=0;
     }
   if (image_info->extract != (char *) NULL)
     {
@@ -918,6 +930,7 @@ static MagickBooleanType WriteJP2Image(const ImageInfo *image_info,Image *image)
       }
       parameters->tcp_numlayers=i+1;
       parameters->cp_fixed_quality=OPJ_TRUE;
+      parameters->cp_disto_alloc=0;
     }
   option=GetImageOption(image_info,"jp2:progression-order");
   if (option != (const char *) NULL)
